from datetime import datetime, timezone

from loguru import logger


class ParserClient:
    """
    Parser Handler Base Class

    Args:
        **kwargs:

    Methods:
        search(self)
        identify_order(self)
        get_order(self)
        replace_instrument(self)

    """

    def __init__(self, **kwargs):
        """
        Initialize the chat client.
        """
        self.name = kwargs.get("name", None)
        self.client = None
        self.enabled = kwargs.get("enabled", None)
        self.action_identifier = kwargs.get("action_identifier", "BUY SELL")
        self.action_identifier = self.action_identifier.lower()
        self.stop_loss_identifier = kwargs.get("stop_loss_identifier", None)
        self.take_profit_identifier = kwargs.get("take_profit_identifier", None)
        self.quantity_identifier = kwargs.get("quantity_identifier", None)
        self.order_type_identifier = kwargs.get("order_type_identifier", None)
        self.leverage_type_identifier = kwargs.get("leverage_type_identifier", None)
        self.comment_identifier = kwargs.get("comment_identifier", None)
        self.stop_loss = kwargs.get("stop_loss", None)
        self.take_profit = kwargs.get("take_profit", None)
        self.quantity = kwargs.get("quantity", None)
        self.instrument_mapping = kwargs.get("instrument_mapping", None)
        self.mapping = kwargs.get("mapping", None)
        self.ignore_instrument = kwargs.get("ignore_instrument", None)

    async def identify_order(
        self,
        my_string: str,
    ) -> dict:
        """
        Identify an order and return a dictionary
        with the order parameters to be implemented in
        the child class

        """

    async def search(self, message: str) -> bool:
        """
        Search an order.

        Args:
            message (str): Message

        Returns:
            bool

        """
        if message:
            order_identifier = message.split()[0].lower()
            # logger.debug("Order identifier: {}", order_identifier)
            # logger.debug("Action identifiers: {}", self.action_identifiers)
            if order_identifier in self.action_identifier:

                # logger.debug("Order identifier found in {}", order_identifier)
                return True

        return False

    async def replace_instrument(self, order):
        """
        Replace instrument by an alternative instrument, if the
        instrument is not in the mapping, it will be ignored.

        Args:
            order (dict):

        Returns:
            dict
        """
        instrument = order["instrument"]
        for item in self.mapping:
            if item["id"] == instrument:
                order["instrument"] = item["alt"]
                break
        logger.debug("Instrument symbol changed", order)
        return order

    async def get_order(
        self,
        msg: str,
    ):
        """
        Get an order from a message. The message can be
        an order or an order identifier

        Args:
            msg (str): Message

        Returns:
            dict

        """
        if not await self.search(msg):
            logger.debug("No order identified")
            return None
        order = await self.identify_order(msg)
        if isinstance(order, dict):
            order["timestamp"] = datetime.now(timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
        if self.instrument_mapping:
            logger.debug("mapping")
            await self.replace_instrument(order)
        if order["instrument"] in self.ignore_instrument:
            logger.debug("Ignoring instrument {}", order["instrument"])
            return
        logger.debug("Order identified {}", order)
        return order



"""
 FindMyOrder Main

"""

import importlib

from loguru import logger

from findmyorder import __version__

from .config import settings


class FindMyOrder:
    """
    Class to find and parse trading order

    Args:
        None

    Returns:
        None

    Methods:
        search(my_string: str) -> bool
        get_info() -> str
        identify_order(my_string: str) -> dict
        get_order(msg: str) -> dict
        replace_instrument(order: dict) -> None

    """

    def __init__(
        self,
    ):
        """
        Initializes the class instance.

        Args:
            self (ClassName): The class instance.

        Returns:
            None
        """

        self.enabled = settings.findmyorder_enabled
        if not self.enabled:
            logger.info("Module is disabled. No Client will be created.")
            return
        self.client_classes = self.get_all_client_classes()
        self.clients = []
        # Create a client for each client in settings.findmyorder
        for name, client_config in settings.findmyorder.items():
            if (
                # Skip empty client configs
                client_config is None
                # Skip non-dict client configs
                or not isinstance(client_config, dict)
                # Skip template and empty string client names
                or name in ["", "template"]
                # Skip disabled clients
                or not client_config.get("enabled")
            ):
                continue

            # Create the client
            logger.debug("Creating client {}", name)
            client = self._create_client(**client_config, name=name)
            # If the client has a valid client attribute, append it to the list
            if client and getattr(client, "client", None):
                self.clients.append(client)

        # Log the number of clients that were created
        logger.info(f"Loaded {len(self.clients)} clients")
        if not self.clients:
            logger.warning(
                "No Client were created. Check your settings or disable the module."
            )

    def _create_client(self, **kwargs):
        """
        Create a client based on the given protocol.

        This function takes in a dictionary of keyword arguments, `kwargs`,
        containing the necessary information to create a client. The required
        key in `kwargs` is "parser_library", which specifies the parser to use
        to identify the order. The value of "parser_library" must match one of the
        libraries supported by findmyorder.

        This function retrieves the class used to create the client based on the
        value of "parser_library" from the mapping of parser names to client classes
        stored in `self.client_classes`. If the value of "parser_library" does not
        match any of the libraries supported, the function logs an error message
        and returns None.

        If the class used to create the client is found, the function creates a
        new instance of the class using the keyword arguments in `kwargs` and
        returns it.

        The function returns a client object based on the specified protocol
        or None if the library is not supported.

        Parameters:
            **kwargs (dict): A dictionary of keyword arguments containing the
            necessary information for creating the client. The required key is
            "parser_library".

        Returns:
            A client object based on the specified protocol or None if the
            library is not supported.

        """
        # library = kwargs.get("parser_library", "standard")
        library = (
            kwargs.get("library")
            or kwargs.get("platform")
            or kwargs.get("protocol")
            or kwargs.get("parser_library")
            or "standard"
        )
        cls = self.client_classes.get((f"{library.capitalize()}Handler"))
        return None if cls is None else cls(**kwargs)

    def get_all_client_classes(self):
        """
        Retrieves all client classes from the `findmyorder.handler` module.

        This function imports the `findmyorder.handler` module and retrieves
        all the classes defined in it.

        The function returns a dictionary where the keys are the
        names of the classes and the values are the corresponding
        class objects.

        Returns:
            dict: A dictionary containing all the client classes
            from the `findmyorder.handler` module.
        """
        provider_module = importlib.import_module("findmyorder.handler")
        return {
            name: cls
            for name, cls in provider_module.__dict__.items()
            if isinstance(cls, type)
        }

    async def get_info(self):
        """
        get info about the class

        Returns:
            str

        """
        version_info = f"ℹ️ {type(self).__name__} {__version__}\n"
        client_info = "".join(
            f"🔎 {client.name}\n🔠 {client.action_identifier}"
            for client in self.clients
        )
        return version_info + client_info.strip()

    async def search(self, message: str) -> bool:
        """
        Search an order.

        Args:
            message (str): Message

        Returns:
            bool

        """
        for client in self.clients:
            logger.debug("Searching with client: {}", client)
            if await client.search(message):
                return True
        return False

    async def identify_order(self, message: str) -> bool:
        """
        Search an order.

        Args:
            message (str): Message

        Returns:
            bool

        """
        for client in self.clients:
            return await client.identify_order(message)

    async def get_order(
        self,
        message: str,
    ):
        """
        Get an order from a message. The message can be
        an order or an order identifier

        Args:
            message (str): Message

        Returns:
            dict

        """
        for client in self.clients:
            return await client.get_order(message)



"""
Standard Parser

"""

from loguru import logger
from pyparsing import (
    Combine,
    Optional,
    Suppress,
    Word,
    alphas,
    nums,
    one_of,
    pyparsing_common,
)

from .handler import ParserClient


class StandardHandler(ParserClient):

    def __init__(self, **kwargs):
        """
        Initialize the Handler object

        """

        super().__init__(**kwargs)
        self.client = "standard"

    async def identify_order(
        self,
        my_string: str,
    ) -> dict:
        """
        Identify an order and return a dictionary
        with the order parameters

        Args:
            my_string (str): Message

        Returns:
            dict

        """
        if not await self.search(my_string):
            logger.debug("No order identified")
            return None
        else:
            action = (
                one_of(self.action_identifier, caseless=True)
                .set_results_name("action")
                .set_parse_action(pyparsing_common.upcase_tokens)
            )
            instrument = Word(alphas + nums).set_results_name("instrument")
            stop_loss = Combine(
                Suppress(self.stop_loss_identifier) + Word(nums)
            ).set_results_name("stop_loss")
            take_profit = Combine(
                Suppress(self.take_profit_identifier) + Word(nums)
            ).set_results_name("take_profit")
            quantity = Combine(
                Suppress(self.quantity_identifier)
                + Word(nums)
                + Optional(Suppress("%"))
            ).set_results_name("quantity")
            order_type = one_of(
                self.order_type_identifier, caseless=True
            ).set_results_name("order_type")
            leverage_type = one_of(
                self.leverage_type_identifier, caseless=True
            ).set_results_name("leverage_type")
            comment = Combine(
                Suppress(self.comment_identifier) + Word(alphas)
            ).set_results_name("comment")

            order_grammar = (
                action("action")
                + Optional(instrument, default=None)
                + Optional(stop_loss, default=self.stop_loss)
                + Optional(take_profit, default=self.take_profit)
                + Optional(quantity, default=self.quantity)
                + Optional(order_type, default=None)
                + Optional(leverage_type, default=None)
                + Optional(comment, default=None)
            )

            order = order_grammar.parse_string(instring=my_string, parse_all=False)
            logger.debug("Order parsed {}", order)
            return order.asDict()



"""
Basic Parser

"""

from pyparsing import (
    Optional,
    Word,
    alphas,
    nums,
    one_of,
    pyparsing_common,
)

from .handler import ParserClient


class BasicHandler(ParserClient):

    def __init__(self, **kwargs):
        """
        Initialize the Handler object

        """

        super().__init__(**kwargs)
        self.client = "basic"

    async def identify_order(
        self,
        my_string: str,
    ) -> dict:
        """
        Identify an order and return a dictionary
        with the order parameters

        Args:
            my_string (str): Message

        Returns:
            dict with the order parameters:
            action, instrument

        """
        action = (
            one_of(self.action_identifier, caseless=True)
            .set_results_name("action")
            .set_parse_action(pyparsing_common.upcase_tokens)
        )
        instrument = Word(alphas + nums).set_results_name("instrument")
        order_grammar = action("action") + Optional(instrument, default=None)
        order = order_grammar.parse_string(instring=my_string, parse_all=False)
        return order.asDict()
